''' Glyph Positioning System (GPOS) tables '''

from __future__ import annotations
from typing import Optional, Union
import logging
from collections import namedtuple

from .fontread import FontReader
from .tables import Coverage, ClassDef, Feature, Script, Language


PlaceMark = namedtuple('PlaceMark', ['dx', 'dy', 'mkmk'])
MarktoBaseAnchor = namedtuple('MarktoBaseAnchor', ['base', 'mark'])


class Gpos:
    ''' Glyph Positioning System Table '''
    def __init__(self, ofst: int, fontfile: FontReader):
        self.ofst = ofst
        self.fontfile = fontfile
        self.language = Language()
        self.fontfile.seek(self.ofst)

        self.vermajor = self.fontfile.readuint16()
        self.verminor = self.fontfile.readuint16()
        scriptofst = self.fontfile.readuint16()
        featureofst = self.fontfile.readuint16()
        lookupofst = self.fontfile.readuint16()
        if self.verminor > 0:
            self.variationofst = self.fontfile.readuint32()
            logging.warning('GPOS has feature variations - unimplemented')

        # Read scripts
        scriptlisttableloc = self.ofst + scriptofst
        scriptcnt = self.fontfile.readuint16(scriptlisttableloc)
        self.scripts = {}
        for i in range(scriptcnt):
            tag = self.fontfile.read(4).decode()
            self.scripts[tag] = Script(
                tag,
                self.fontfile.readuint16() + scriptlisttableloc,
                self.fontfile)

        # Read features
        featurelisttableloc = self.ofst + featureofst
        featurecnt = self.fontfile.readuint16(featurelisttableloc)
        featurelist = []
        for i in range(featurecnt):
            featurelist.append(Feature(
                self.fontfile.read(4).decode(),
                self.fontfile.readuint16() + featurelisttableloc,
                self.fontfile))

        # Read Lookups
        lookuplisttableloc = self.ofst + lookupofst
        lookupcnt = self.fontfile.readuint16(lookuplisttableloc)
        self.lookups = []
        for i in range(lookupcnt):
            self.lookups.append(GposLookup(
                self.fontfile.readuint16() + lookuplisttableloc,
                self.fontfile))

        # Put everything in a dictionary for access
        self.features = {}
        for scrname, script in self.scripts.items():
            langdict = {}
            for langname, lang in script.languages.items():
                langfeatures = [featurelist[i] for i in lang.featureidxs]
                featnames = [f.tag for f in langfeatures]
                featdict = {}
                for feat in featnames:
                    lookups = langfeatures[featnames.index(feat)].lookupids
                    tables = [self.lookups[i] for i in lookups]
                    featdict[feat] = tables
                langdict[langname] = featdict
            self.features[scrname] = langdict

    def features_active(self):
        ''' Dictionary of features active in the current script/language system '''
        return self.features.get(self.language.script, {}).get(self.language.language, {})

    def kern(self, glyph1: int, glyph2: int) -> tuple[dict, dict]:
        ''' Get kerning adjustmnet for glyph1 and glyph2 '''
        feattable = self.features_active()
        if 'kern' in feattable:
            for table in feattable['kern']:
                for subtable in table.subtables:
                    v1, v2 = subtable.get_adjust(glyph1, glyph2)
                    if v1 or v2:
                        return v1, v2
        return {}, {}  # No adjustments

    def placemark(self, base: int, mark: int) -> Optional[PlaceMark]:
        ''' Return dx, dy postition wrt original mark position '''
        features = self.features_active()
        for feat in ['mark', 'mkmk']:
            if feat in features:
                for lookup in features[feat]:
                    for subtable in lookup.subtables:
                        if not hasattr(subtable, 'anchor'):
                            logging.debug('skipping unimplemented placemark subtable %s', subtable)
                            continue

                        anchors = subtable.anchor(base, mark)
                        if anchors is not None:
                            dx = anchors.base.x - anchors.mark.x
                            dy = anchors.base.y - anchors.mark.y
                            logging.debug('Positioning Mark %s on %s: (%s, %s)',
                                          mark, base, dx, dy)
                            mkmk = isinstance(subtable, MarkToMarkSubtable)
                            return PlaceMark(dx, dy, mkmk)   # Use first one found
        return None

    def __repr__(self):
        return f'<GPOS Table v{self.vermajor}.{self.verminor}>'


class GposLookup:
    ''' GPOS Lookup Table '''
    def __init__(self, ofst: int, fontfile: FontReader):
        self.ofst = ofst
        self.fontfile = fontfile
        fileptr = self.fontfile.tell()

        RIGHT_TO_LEFT = 0x0001
        IGNORE_BASE_GLYPHS = 0x0002
        IGNORE_LIGATURES = 0x0004
        IGNORE_MARKS = 0x0008
        USE_MARK_FILTERING_SET = 0x0010
        MARK_ATTACHMENT_TYPE_MASK = 0xFF00

        self.fontfile.seek(self.ofst)
        self.type = self.fontfile.readuint16()
        self.flag = self.fontfile.readuint16()
        subtablecnt = self.fontfile.readuint16()
        self.tableofsts = []
        for i in range(subtablecnt):
            self.tableofsts.append(self.fontfile.readuint16())
        self.markfilterset = None
        if self.flag & USE_MARK_FILTERING_SET:
            self.markfilterset = self.fontfile.readuint16()

        self.subtables: list[Union[PairAdjustmentSubtable,
                                   MarkToBaseSubtable,
                                   MarkToMarkSubtable]] = []
        for tblofst in self.tableofsts:
            tabletype = self.type
            if self.type == 9:  # Extension table - converts to another type
                fmt = self.fontfile.readuint16(self.ofst + tblofst)
                assert fmt == 1
                tabletype = self.fontfile.readuint16()
                tblofst += self.fontfile.readuint32()

            if tabletype == 2:  # Pair adjustment positioning
                self.subtables.append(PairAdjustmentSubtable(
                        tblofst + self.ofst, self.fontfile))
            elif tabletype == 4:  # Mark-to-base
                self.subtables.append(MarkToBaseSubtable(
                        tblofst + self.ofst, self.fontfile))
            elif tabletype == 6:  # Mark-to-mark
                self.subtables.append(MarkToMarkSubtable(
                        tblofst + self.ofst, self.fontfile))
            else:
                logging.debug('Unimplemented GPOS Lookup Type %s', self.type)

        self.fontfile.seek(fileptr)  # Put file pointer back

    def __repr__(self):
        return f'<GPOSLookup Type {self.type} {hex(self.ofst)}>'


class PairAdjustmentSubtable:
    ''' Pair Adjustment Table (GPOS Lookup Type 2)
        Informs kerning between pairs of glyphs
    '''
    def __init__(self, ofst: int, fontfile: FontReader):
        self.ofst = ofst
        self.fontfile = fontfile
        fileptr = self.fontfile.tell()

        self.fontfile.seek(self.ofst)
        self.posformat = self.fontfile.readuint16()
        self.covofst = self.fontfile.readuint16()
        self.valueformat1 = self.fontfile.readuint16()
        self.valueformat2 = self.fontfile.readuint16()

        if self.posformat == 1:
            pairsetcount = self.fontfile.readuint16()
            pairsetofsts = []
            for i in range(pairsetcount):
                pairsetofsts.append(self.fontfile.readuint16())

            PairValue = namedtuple('PairValue', ['secondglyph', 'value1', 'value2'])
            
            self.pairsets = []
            for i in range(pairsetcount):
                self.fontfile.seek(pairsetofsts[i] + self.ofst)
                paircnt = self.fontfile.readuint16()
                pairs = []
                for p in range(paircnt):
                    pairs.append(PairValue(
                        self.fontfile.readuint16(),
                        self.fontfile.readvaluerecord(self.valueformat1),
                        self.fontfile.readvaluerecord(self.valueformat2)))
                self.pairsets.append(pairs)

        elif self.posformat == 2:
            classdef1ofst = self.fontfile.readuint16()
            classdef2ofst = self.fontfile.readuint16()
            class1cnt = self.fontfile.readuint16()
            class2cnt = self.fontfile.readuint16()

            self.classrecords = []
            for i in range(class1cnt):
                class2recs = []
                for j in range(class2cnt):
                    class2recs.append(
                        (self.fontfile.readvaluerecord(self.valueformat1),
                         self.fontfile.readvaluerecord(self.valueformat2)))
                self.classrecords.append(class2recs)

            self.class1def = ClassDef(self.ofst + classdef1ofst, self.fontfile)
            self.class2def = ClassDef(self.ofst + classdef2ofst, self.fontfile)

        else:
            raise ValueError('Invalid posformat in PairAdjustmentSubtable')

        self.coverage = Coverage(self.covofst+self.ofst, self.fontfile)
        self.fontfile.seek(fileptr)  # Put file pointer back

    def get_adjust(self, glyph1: int, glyph2: int) -> tuple[Optional[dict], Optional[dict]]:
        ''' Get kerning adjustment for glyph1 and glyph2 pair '''
        v1 = v2 = None

        # Look up first glyph in coverage table
        covidx = self.coverage.covidx(glyph1)
        if covidx is not None:

            # Look up second glyph
            if self.posformat == 1:
                for p in self.pairsets[covidx]:
                    if p.secondglyph == glyph2:
                        v1 = p.value1
                        v2 = p.value2
                        break

            else:
                c1 = self.class1def.get_class(glyph1)
                c2 = self.class2def.get_class(glyph2)
                if c1 is not None and c2 is not None:
                    v1, v2 = self.classrecords[c1][c2]

        return v1, v2

    def __repr__(self):
        return f'<PairAdjustmentSubtable {hex(self.ofst)}>'


def read_markarray_table(ofst, fontfile):
    ''' Read MarkArray table from font file '''
    cnt = fontfile.readuint16(ofst)
    MarkRecord = namedtuple('MarkRecord', ['markclass', 'anchortable'])
    markrecords = []
    for i in range(cnt):
        markclass = fontfile.readuint16()
        anchorofst = fontfile.readuint16()
        ptr = fontfile.tell()
        anchortable = read_anchortable(ofst+anchorofst, fontfile)
        markrecords.append(MarkRecord(markclass, anchortable))
        ptr = fontfile.seek(ptr)
    return markrecords


def read_anchortable(ofst, fontfile):
    ''' Read anchor table from font file '''
    Anchor = namedtuple('Anchor', ['x', 'y', 'anchorpoint', 'xofst', 'yofst'])
    fmt = fontfile.readuint16(ofst)
    point = None
    xofst = yofst = None
    x = fontfile.readint16()
    y = fontfile.readint16()
    if fmt == 2:
        point = fontfile.readuint16()
    elif fmt == 3:
        xofst = fontfile.readuint16()
        yofst = fontfile.readuint16()
        logging.warning('Anchor references Device Table - unimplemented')
    return Anchor(x, y, point, xofst, yofst)


def read_basearray(ofst, fontfile, markclasscount):
    ''' Read BaseArray table used by Mark-to-Base Lookup '''
    basecnt = fontfile.readuint16(ofst)
    basearray = []
    for i in range(basecnt):
        anchortables = []
        for j in range(markclasscount):
            baseanchorofst = fontfile.readuint16()
            ptr = fontfile.tell()
            if baseanchorofst == 0:
                anchortables.append([])
            else:
                anchortables.append(read_anchortable(
                    ofst+baseanchorofst, fontfile))
            fontfile.seek(ptr)
        basearray.append(anchortables)
    return basearray


class MarkToBaseSubtable:
    ''' Mark-To-Base Positioning Table (GPOS Lookup Type 4) '''
    def __init__(self, ofst: int, fontfile: FontReader):
        self.ofst = ofst
        self.fontfile = fontfile

        self.fmt = self.fontfile.readuint16(self.ofst)
        assert self.fmt == 1
        markcovofst = self.fontfile.readuint16()
        basecovofst = self.fontfile.readuint16()
        markclasscnt = self.fontfile.readuint16()
        markarrayofst = self.fontfile.readuint16()
        basearrayofst = self.fontfile.readuint16()

        self.markarray = read_markarray_table(self.ofst+markarrayofst, self.fontfile)
        self.basearray = read_basearray(self.ofst+basearrayofst, self.fontfile, markclasscnt)
        self.markcoverage = Coverage(self.ofst+markcovofst, self.fontfile)
        self.basecoverage = Coverage(self.ofst+basecovofst, self.fontfile)

    def anchor(self, base: int, mark: int) -> Optional[MarktoBaseAnchor]:
        ''' Get anchors for mark glyph with respect to base glyph '''
        markid = self.markcoverage.covidx(mark)
        if markid is None:
            return None

        baseid = self.basecoverage.covidx(base)
        if baseid is None:
            return None

        markanchor = self.markarray[markid]
        baseanchor = self.basearray[baseid][markanchor.markclass]
        return MarktoBaseAnchor(baseanchor, markanchor.anchortable)


class MarkToMarkSubtable:
    ''' Mark-To-Mark Positioning Table (GPOS Lookup Type 6) '''
    def __init__(self, ofst: int, fontfile: FontReader):
        self.ofst = ofst
        self.fontfile = fontfile

        self.fmt = self.fontfile.readuint16(self.ofst)
        assert self.fmt == 1
        mark1covofst = self.fontfile.readuint16()
        mark2covofst = self.fontfile.readuint16()
        markclasscnt = self.fontfile.readuint16()
        mark1arrayofst = self.fontfile.readuint16()
        mark2arrayofst = self.fontfile.readuint16()

        mark2count = self.fontfile.readuint16(self.ofst+mark2arrayofst)
        self.mark2array = []
        for i in range(mark2count):
            anchorofsts = []
            for j in range(markclasscnt):
                anchorofsts.append(self.fontfile.readuint16())
            anchortables = []
            ptr = self.fontfile.tell()
            for anchorofst in anchorofsts:
                anchortables.append(read_anchortable(
                    self.ofst+mark2arrayofst+anchorofst,
                    self.fontfile))
            self.mark2array.append(anchortables)
            self.mark1array = read_markarray_table(
                self.ofst+mark1arrayofst, self.fontfile)
            self.fontfile.seek(ptr)

        self.mark1coverage = Coverage(
            self.ofst+mark1covofst, self.fontfile)
        self.mark2coverage = Coverage(
            self.ofst+mark2covofst, self.fontfile)

    def anchor(self, mark2: int, mark1: int) -> Optional[MarktoBaseAnchor]:
        ''' Get anchors for mark glyph with respect to first mark glyph '''
        mark1id = self.mark1coverage.covidx(mark1)
        if mark1id is None:
            return None

        mark2id = self.mark2coverage.covidx(mark2)
        if mark2id is None:
            return None

        mark1anchor = self.mark1array[mark1id]
        mark2anchor = self.mark2array[mark2id][mark1anchor.markclass]
        return MarktoBaseAnchor(mark2anchor, mark1anchor.anchortable)



''' Glyph classes '''

from __future__ import annotations
from typing import Sequence, Optional, TYPE_CHECKING
from types import SimpleNamespace

import os
import xml.etree.ElementTree as ET

from .fonttypes import GlyphComp, BBox
from .config import config
from .svgpath import fmt, SVGOpType
from .glyphinspect import InspectGlyph, DescribeGlyph

if TYPE_CHECKING:
    from .font import Font


class SimpleGlyph:
    ''' Simple Glyph '''
    DFLT_SIZE_PT = 12   # Draw <symbols> in this point size

    def __init__(self, index: int, operators: Sequence[SVGOpType],
                 bbox: BBox, font: Font):
        self.index = index
        self.operators = operators
        self.bbox = bbox
        self.path = SimpleNamespace()
        self.path.bbox = bbox  # Only for backward-compatibility
        self.font = font
        basename, _ = os.path.splitext(os.path.basename(self.font.info.filename))
        basename = ''.join(c for c in basename if c.isalnum())
        self.id = f'{basename}_{index}'
        self._points_per_unit = self.DFLT_SIZE_PT / self.font.info.layout.unitsperem

    def _repr_svg_(self):
        return ET.tostring(self.svgxml(), encoding='unicode')

    @property
    def char(self) -> set[str]:
        ''' Get set of unicode character represented by this glyph '''
        if self.font.cmap:
            return self.font.cmap.char(self.index)
        return set()

    def funits_to_points(self, value: float, scale_factor: float = 1) -> float:
        ''' Convert font units to points '''
        return value * self._points_per_unit * scale_factor

    def place(self, x: float, y: float, point_size: float) -> Optional[ET.Element]:
        ''' Get <use> svg tag translated/scaled to the right position '''
        scale_factor = point_size / self.DFLT_SIZE_PT
        yshift = self.funits_to_points(
            max(self.bbox.ymax, self.font.info.layout.ymax), scale_factor)
        elm: Optional[ET.Element]
        if config.svg2:
            elm = ET.Element('use')
            elm.attrib['href'] = f'#{self.id}'
            dx = min(self.funits_to_points(self.bbox.xmin, scale_factor), 0)
            elm.attrib['transform'] = (f'translate({fmt(x+dx)} '
                                       f'{fmt(y-yshift)}) scale({fmt(scale_factor)})')
        else:
            elm = self.svgpath(x0=x, y0=y, scale_factor=scale_factor)
        return elm

    def advance(self, nextchr=None) -> float:
        ''' Get advance width in glyph units, including kerning if nextchr is defined '''
        if nextchr:
            nextchr = nextchr.index
        return self.font.advance(self.index, nextchr)

    def svgpath(self, x0: float = 0, y0: float = 0, scale_factor: float = 1) -> Optional[ET.Element]:
        ''' Get svg <path> element for glyph, normalized to 12-point font '''
        path = ''
        for operator in self.operators:
            segment = operator.path(x0, y0, scale=self.funits_to_points(1, scale_factor))
            if segment[0] == 'M' and path != '':
                path += 'Z '  # Close intermediate segments
            path += segment
        if path == '':
            return None  # Don't return empty path
        path += 'Z '
        return ET.Element('path', attrib={'d': path})

    def svgsymbol(self) -> ET.Element:
        ''' Get svg <symbol> element for this glyph, scaled to 12-point font '''
        xmin = min(self.funits_to_points(self.bbox.xmin), 0)
        xmax = self.funits_to_points(self.bbox.xmax)
        width = xmax-xmin
        ymax = self.funits_to_points(max(self.font.info.layout.ymax, self.bbox.ymax))
        ymin = self.funits_to_points(min(self.font.info.layout.ymin, self.bbox.ymin))
        height = ymax - ymin

        sym = ET.Element('symbol')
        sym.attrib['id'] = self.id
        sym.attrib['width'] = fmt(width)
        sym.attrib['height'] = fmt(height)
        sym.attrib['viewBox'] = f'{fmt(xmin)} {fmt(-ymax)} {fmt(width)} {fmt(height)}'
        if (path := self.svgpath()) is not None:
            sym.append(path)
        return sym

    def svg(self, point_size: Optional[float] = None) -> str:
        ''' Get SVG as string '''
        return ET.tostring(self.svgxml(point_size), encoding='unicode')

    def svgxml(self, point_size: Optional[float] = None) -> ET.Element:
        ''' Standalong SVG '''
        point_size = point_size if point_size else config.fontsize
        scale_factor = point_size / self.DFLT_SIZE_PT

        # Width varies by character, but height is constant for the whole font
        # View should include whole character, even if it goes negative/outside the advwidth
        xmin = self.funits_to_points(min(self.bbox.xmin, 0), scale_factor)
        xmax = self.funits_to_points(self.bbox.xmax, scale_factor)
        ymin = self.funits_to_points(min(self.bbox.ymin, self.font.info.layout.ymin), scale_factor)

        # ymax can go above font's ymax for extended (ie math) glyphs
        ymax = self.funits_to_points(max(self.bbox.ymax, self.font.info.layout.ymax), scale_factor)
        width = xmax - xmin
        height = ymax - ymin
        base = ymax

        svg = ET.Element('svg')
        svg.attrib['width'] = fmt(width)
        svg.attrib['height'] = fmt(height)
        svg.attrib['xmlns'] = 'http://www.w3.org/2000/svg'
        svg.attrib['viewBox'] = f'{fmt(xmin)} 0 {fmt(width)} {fmt(height)}'
        if not config.svg2:
            svg.attrib['xmlns:xlink'] = 'http://www.w3.org/1999/xlink'
            elm = self.svgpath(x0=0, y0=base, scale_factor=scale_factor)
            if elm is not None:
                svg.append(elm)
        else:
            symbol = self.svgsymbol()
            svg.append(symbol)
            g = ET.SubElement(svg, 'use')
            g.attrib['href'] = f'#{self.id}'
            g.attrib['transform'] = (f'translate({fmt(xmin)}, '
                                     f'{fmt(base-ymax)}) scale({fmt(scale_factor)})')
        return svg

    def test(self, pxwidth: float = 400, pxheight: float = 400) -> InspectGlyph:
        ''' Get Glyph Test representation showing vertices and borders '''
        return InspectGlyph(self, pxwidth, pxheight)

    def describe(self) -> DescribeGlyph:
        ''' Get Glyph Test representation showing vertices and borders '''
        return DescribeGlyph(self)


class CompoundGlyph(SimpleGlyph):
    ''' Compound glyph, made of multiple other Glyphs '''
    def __init__(self, index: int, glyphs: GlyphComp, font: Font):
        self.index = index
        self.glyphs = glyphs
        operators = self._buildcompound()
        super().__init__(index, operators, self.glyphs.bbox, font)

    def _buildcompound(self) -> list[SVGOpType]:
        ''' Combine multiple glyphs into one set of contours '''
        xoperators = []
        for glyph, xform in zip(self.glyphs.glyphs, self.glyphs.xforms):
            if xform.match:
                raise NotImplementedError('Compound glyph match transform')

            m0 = max(abs(xform.a), abs(xform.b))
            n0 = max(abs(xform.c), abs(xform.d))
            m = 2*m0 if abs(abs(xform.a)-abs(xform.c)) <= 33/65536 else m0
            n = 2*n0 if abs(abs(xform.b)-abs(xform.d)) <= 33/65536 else n0
            for operator in glyph.operators:
                xoperators.append(operator.xform(xform.a, xform.b, xform.c,
                                                 xform.d, xform.e, xform.f, m, n))
        return xoperators


class EmptyGlyph(SimpleGlyph):
    ''' Glyph with no contours (like a space) '''
    def __init__(self, index: int, font: Font):
        super().__init__(index, [], BBox(0, 0, 0, 0), font)



''' Glyph Substitution (GSUB) tables '''

from __future__ import annotations
from typing import Optional, TYPE_CHECKING
from collections import namedtuple
import random
import logging

from .tables import Script, Feature, Coverage, Language, ClassDef
from .fonttypes import FontFeatures

if TYPE_CHECKING:
    from .fontread import FontReader


ChainedSeqRule = namedtuple('ChainedSeqRule', ['backtrack', 'input', 'lookahead', 'sequence'])
SequenceLookupRecord = namedtuple('SequenceLookupRecord', ['sequenceindex', 'lookupindex'])


class LookupSubtable:
    ''' Generic/unimplemented GSUB Lookup Sub Table '''
    def __init__(self, ofst: int, fontfile: FontReader):
        self.ofst = ofst
        self.fontfile = fontfile

    def sub(self, glyphids: list[int], lookups: list[GSUBLookup], name: Optional[str] = None) -> list[int]:
        ''' Apply glyph substitution to list of glyph ids '''
        logging.debug('Glyph substitution from unimplemented GSUB subtable type %s', self.ofst)
        return glyphids


class LookupSingleSub1(LookupSubtable):
    ''' GSUB Single Glyph Substitution  (LookupType 1, format 1) '''
    def __init__(self, ofst: int, fontfile: FontReader):
        super().__init__(ofst, fontfile)
        self.fmt = fontfile.readuint16(ofst)
        assert self.fmt == 1
        covofst = fontfile.readuint16()
        self.deltagid = fontfile.readint16()
        self.covtable = Coverage(ofst + covofst, fontfile)

    def sub(self, glyphids: list[int], lookups: list[GSUBLookup], name: Optional[str] = None) -> list[int]:
        ''' Apply glyph substitution to list of glyph ids '''
        for i in range(len(glyphids)):
            covidx = self.covtable.covidx(glyphids[i])
            if covidx is not None:
                glyphids[i] += self.deltagid
        return glyphids


class LookupSingleSub2(LookupSubtable):
    ''' GSUB Single Glyph Substitution  (LookupType 1, format 2) '''
    def __init__(self, ofst: int, fontfile: FontReader):
        super().__init__(ofst, fontfile)
        self.fmt = fontfile.readuint16(ofst)
        assert self.fmt == 2
        covofst = fontfile.readuint16()
        glyphcount = fontfile.readuint16()
        self.subglyphids = []
        for _ in range(glyphcount):
            self.subglyphids.append(fontfile.readuint16())
        self.covtable = Coverage(ofst + covofst, fontfile)

    def sub(self, glyphids: list[int], lookups: list[GSUBLookup], name: Optional[str]=None) -> list[int]:
        ''' Apply glyph substitution to list of glyph ids '''
        for i in range(len(glyphids)):
            covidx = self.covtable.covidx(glyphids[i])
            if covidx is not None:
                glyphids[i] = self.subglyphids[covidx]
        return glyphids


class LookupMultipleSub(LookupSubtable):
    ''' GSUB Multiple Glyph Substitution  (LookupType 2) '''
    def __init__(self, ofst: int, fontfile: FontReader):
        super().__init__(ofst, fontfile)
        self.fmt = fontfile.readuint16(ofst)
        assert self.fmt == 1
        covofst = fontfile.readuint16()
        seqcount = fontfile.readuint16()
        seqofsts = []
        for _ in range(seqcount):
            seqofsts.append(fontfile.readuint16())
        self.subglyphids = []
        for seqofst in seqofsts:
            glyphcnt = fontfile.readuint16(ofst+seqofst)
            subglyphs = []
            for _ in range(glyphcnt):
                subglyphs.append(fontfile.readuint16())
            self.subglyphids.append(subglyphs)

        self.covtable = Coverage(ofst + covofst, fontfile)

    def sub(self, glyphids: list[int], lookups: list[GSUBLookup], name: Optional[str] = None) -> list[int]:
        ''' Apply glyph substitution to list of glyph ids '''
        newglyphids = []
        for gid in glyphids:
            covidx = self.covtable.covidx(gid)
            if covidx is not None:
                newglyphids.extend(self.subglyphids[covidx])
            else:
                newglyphids.append(gid)
        return newglyphids


class LookupAlternate(LookupSubtable):
    ''' GSUB Alternate Glyph Substitution  (LookupType 3) '''
    def __init__(self, ofst: int, fontfile: FontReader):
        super().__init__(ofst, fontfile)
        self.fmt = fontfile.readuint16(ofst)
        assert self.fmt == 1
        covofst = fontfile.readuint16()
        cnt = fontfile.readuint16()
        altofsts = []
        for i in range(cnt):
            altofsts.append(fontfile.readuint16())
        self.altglyphs = []
        for altofst in altofsts:
            cnt = self.fontfile.readuint16(ofst+altofst)
            gids = []
            for i in range(cnt):
                gids.append(self.fontfile.readuint16())
            self.altglyphs.append(gids)
        self.covtable = Coverage(ofst + covofst, fontfile)

    def sub(self, glyphids: list[int], lookups: list[GSUBLookup], name: Optional[str] = None) -> list[int]:
        ''' Apply glyph substitution to list of glyph ids '''
        for i in range(len(glyphids)):
            covidx = self.covtable.covidx(glyphids[i])
            if covidx is not None:

                if name == 'rand':
                    glyphids[i] = random.choice(self.altglyphs[covidx])
                else:
                    altgids = self.altglyphs[covidx]
                    # TODO - user choice! return first one for now.
                    glyphids[i] = altgids[0]

        return glyphids


class LookupLigatureSub(LookupSubtable):
    ''' Ligature Substitution Subtable (LookupType 4) '''
    def __init__(self, ofst: int, fontfile: FontReader):
        super().__init__(ofst, fontfile)
        self.ofst = ofst
        self.fontfile = fontfile
        fmt = fontfile.readuint16(self.ofst)
        assert fmt == 1
        covofst = self.fontfile.readuint16()
        cnt = self.fontfile.readuint16()
        ligsetofsts = []
        for _ in range(cnt):
            ligsetofsts.append(self.fontfile.readuint16())

        self.ligsets = []
        for ofst in ligsetofsts:
            ofst = ofst + self.ofst
            cnt = self.fontfile.readuint16(ofst)
            ligofsts = []
            for _ in range(cnt):
                ligofsts.append(self.fontfile.readuint16())

            ligs = {}
            for _ in range(cnt):
                ligglyph = self.fontfile.readuint16()
                compcount = self.fontfile.readuint16()
                compglyphs = []
                for _ in range(compcount-1):
                    compglyphs.append(self.fontfile.readuint16())
                ligs[tuple(compglyphs)] = ligglyph
            self.ligsets.append(ligs)
        self.covtable = Coverage(self.ofst+covofst, self.fontfile)

    def sub(self, glyphids: list[int], lookups: list[GSUBLookup], name: Optional[str] = None) -> list[int]:
        ''' Apply glyph substitution to list of glyph ids '''
        newgids = []
        i = 0
        while i < len(glyphids):
            gid = glyphids[i]
            covidx = self.covtable.covidx(gid)
            if covidx is not None:
                ligset = self.ligsets[covidx]
                for ls, subglyph in ligset.items():
                    if glyphids[i+1:i+1+len(ls)] == list(ls):
                        newgids.append(subglyph)
                        i += len(ls)+1
                        break
                else:  # nobreak
                    newgids.append(gid)
                    i += 1
            else:
                newgids.append(gid)
                i += 1
        return newgids


class LookupChainedSub3(LookupSubtable):
    ''' Chained Contexts Subtable (LookupType 6, format 3) '''
    def __init__(self, ofst: int, fontfile: FontReader):
        super().__init__(ofst, fontfile)
        self.ofst = ofst
        self.fontfile = fontfile
        self.fmt = self.fontfile.readuint16(self.ofst)
        assert self.fmt == 3

        backtrackcnt = self.fontfile.readuint16()
        backofsts = []
        for i in range(backtrackcnt):
            backofsts.append(self.fontfile.readuint16())
        inptcnt = self.fontfile.readuint16()
        inptofsts = []
        for i in range(inptcnt):
            inptofsts.append(self.fontfile.readuint16())
        lookcnt = self.fontfile.readuint16()
        lookofsts = []
        for i in range(lookcnt):
            lookofsts.append(self.fontfile.readuint16())
        seqlookupcnt = self.fontfile.readuint16()
        self.seqlookups = []
        for i in range(seqlookupcnt):
            self.seqlookups.append(SequenceLookupRecord(
                self.fontfile.readuint16(),
                self.fontfile.readuint16()))

        self.backCoverage = []
        for bofst in backofsts:
            self.backCoverage.append(Coverage(self.ofst+bofst, self.fontfile))
        self.inptCoverage = []
        for iofst in inptofsts:
            self.inptCoverage.append(Coverage(self.ofst+iofst, self.fontfile))
        self.lookaheadCoverage = []
        for lofst in lookofsts:
            self.lookaheadCoverage.append(Coverage(self.ofst+lofst, self.fontfile))

    def sub(self, glyphids: list[int], lookups: list[GSUBLookup], name: Optional[str] = None) -> list[int]:
        ''' Apply glyph substitution to list of glyph ids '''
        ilen = len(self.inptCoverage)
        i = len(self.backCoverage)
        while i < len(glyphids) - len(self.lookaheadCoverage) - ilen + 1:
            covidx = [cov.covidx(glyphids[i+k]) for k, cov in enumerate(self.inptCoverage)]
            if None in covidx:
                i += 1
                continue
            covidx = [cov.covidx(glyphids[i-k-1]) for k, cov in enumerate(self.backCoverage)]
            if None in covidx:
                i += 1
                continue
            covidx = [cov.covidx(glyphids[i+ilen+k]) for k, cov in enumerate(self.lookaheadCoverage)]
            if None in covidx:
                i += 1
                continue

            # match!
            for seqlookup in self.seqlookups:
                glyphids[i+seqlookup.sequenceindex:i+1+ilen] = \
                    lookups[seqlookup.lookupindex].sub(glyphids[i:i+1+ilen], lookups)

            i += len(self.inptCoverage)

        return glyphids


class LookupChainedSub2(LookupSubtable):
    ''' Chained Contexts Subtable (LookupType 6, format 2) '''
    def __init__(self, ofst: int, fontfile: FontReader):
        super().__init__(ofst, fontfile)
        self.ofst = ofst
        self.fontfile = fontfile
        self.fmt = fontfile.readuint16(self.ofst)
        assert self.fmt == 2
        covofst = self.fontfile.readuint16()
        backtrackofst = self.fontfile.readuint16()
        inputofst = self.fontfile.readuint16()
        lookaheadofst = self.fontfile.readuint16()
        rulesetcnt = self.fontfile.readuint16()
        rulesetofsts = []
        for i in range(rulesetcnt):
            rulesetofsts.append(self.fontfile.readuint16())

        self.rules = []
        for rulesetofst in rulesetofsts:
            rulecnt = self.fontfile.readuint16(self.ofst+rulesetofst)
            ruleofsts = []
            for i in range(rulecnt):
                ruleofsts.append(self.fontfile.readuint16())

            ruleset = []
            for ruleofst in ruleofsts:
                backglyphcnt = self.fontfile.readuint16(self.ofst+rulesetofst+ruleofst)
                backseq = []
                for i in range(backglyphcnt):
                    backseq.append(self.fontfile.readuint16())
                inglyphcnt = self.fontfile.readuint16()
                inputseq = []
                for i in range(inglyphcnt-1):
                    inputseq.append(self.fontfile.readuint16())
                lookaheadglyphcnt = self.fontfile.readuint16()
                lookaheadseq = []
                for i in range(lookaheadglyphcnt):
                    lookaheadseq.append(self.fontfile.readuint16())
                seqlookupcnt = self.fontfile.readuint16()
                sequencelookup = []
                for i in range(seqlookupcnt):
                    sequencelookup.append(
                        SequenceLookupRecord(
                            self.fontfile.readuint16(),
                            self.fontfile.readuint16()))
                ruleset.append(
                    ChainedSeqRule(backseq, inputseq, lookaheadseq, sequencelookup))
            self.rules.append(ruleset)
        self.covtable = Coverage(self.ofst+covofst, self.fontfile)
        self.backtrackClass = ClassDef(self.ofst+backtrackofst, self.fontfile)
        self.inputClass = ClassDef(self.ofst+inputofst, self.fontfile)
        self.lookaheadClass = ClassDef(self.ofst+lookaheadofst, self.fontfile)

    def sub(self, glyphids: list[int], lookups: list[GSUBLookup], name: Optional[str] = None) -> list[int]:
        ''' Apply glyph substitution to list of glyph ids '''
        i = 0
        while i < len(glyphids):
            gid = glyphids[i]
            covidx = self.covtable.covidx(gid)
            if covidx is not None:
                classid = self.inputClass.get_class(gid)
                if classid is not None:
                    ruleset = self.rules[classid]
                    for rule in ruleset:
                        inpt = [self.inputClass.get_class(g)
                                for g in glyphids[i+1:i+1+len(rule.input)]]
                        if inpt != rule.input:
                            continue
                        backtrack = [self.backtrackClass.get_class(g)
                                     for g in glyphids[i-len(rule.backtrack):i]]
                        if backtrack[::-1] != rule.backtrack:  # Rule is reverse order
                            continue
                        aheadlen = len(rule.input) + len(rule.lookahead)
                        lookahead = [self.lookaheadClass.get_class(g)
                                     for g in glyphids[i+1+len(rule.input):i+1+aheadlen]]

                        if lookahead != rule.lookahead:
                            continue

                        # Match
                        for seqlookup in rule.sequence:
                            glyphids[i+seqlookup.sequenceindex:i+1+len(rule.input)] = \
                                lookups[seqlookup.lookupindex].sub(glyphids[i:i+1+len(rule.input)], lookups)
                i += 1
            else:
                i += 1
        return glyphids


class LookupChainedSub1(LookupSubtable):
    ''' Chained Contexts Subtable (LookupType 6, format 1) '''
    def __init__(self, ofst: int, fontfile: FontReader):
        super().__init__(ofst, fontfile)
        self.ofst = ofst
        self.fontfile = fontfile
        self.fmt = fontfile.readuint16(self.ofst)
        assert self.fmt == 1
        covofst = self.fontfile.readuint16()
        cnt = self.fontfile.readuint16()
        setofsts = []
        for i in range(cnt):
            setofsts.append(self.fontfile.readuint16())

        self.rules = []
        for setofst in setofsts:
            cnt = self.fontfile.readuint16(self.ofst+setofst)
            ruleofsts = []
            for i in range(cnt):
                ruleofsts.append(self.fontfile.readuint16())
            ruleset = []
            for ruleofst in ruleofsts:
                backglyphcount = self.fontfile.readuint16(self.ofst+setofst+ruleofst)
                backtrackSeq = []
                for i in range(backglyphcount):
                    backtrackSeq.append(self.fontfile.readuint16())
                inputglyphcount = self.fontfile.readuint16()
                inputSequence = []
                for i in range(inputglyphcount-1):
                    inputSequence.append(self.fontfile.readuint16())
                lookaheadglyphcount = self.fontfile.readuint16()
                lookaheadSequence = []
                for i in range(lookaheadglyphcount):
                    lookaheadSequence.append(self.fontfile.readuint16())
                seqlookupcnt = self.fontfile.readuint16()
                sequencelookup = []
                for i in range(seqlookupcnt):
                    sequencelookup.append(SequenceLookupRecord(
                        self.fontfile.readuint16(),   # SequenceIndex
                        self.fontfile.readuint16()))  # LookupListIndex
                ruleset.append(ChainedSeqRule(
                    backtrackSeq, inputSequence, lookaheadSequence, sequencelookup))
            self.rules.append(ruleset)
        self.covtable = Coverage(self.ofst+covofst, self.fontfile)

    def sub(self, glyphids: list[int], lookups: list[GSUBLookup], name: Optional[str] = None) -> list[int]:
        ''' Apply glyph substitution to list of glyph ids '''
        i = 0
        while i < len(glyphids):
            gid = glyphids[i]
            covidx = self.covtable.covidx(gid)
            if covidx is not None:
                ruleset = self.rules[covidx]
                for rule in ruleset:
                    inpt = glyphids[i+1:i+1+len(rule.input)]
                    if inpt != rule.input:
                        i += 1
                        continue
                    backtrack = glyphids[i-len(rule.backtrack):i]
                    if backtrack[::-1] != rule.backtrack:  # Rule is reverse order
                        i += 1
                        continue
                    aheadlen = len(rule.input) + len(rule.lookahead)
                    lookahead = glyphids[i+1+len(rule.input):i+1+aheadlen]

                    if lookahead != rule.lookahead:
                        i += 1
                        continue

                    # Match
                    for seqlookup in rule.sequence:
                        glyphids[i+seqlookup.sequenceindex:i+1+len(rule.input)] = \
                            lookups[seqlookup.lookupindex].sub(glyphids[i:i+1+len(rule.input)], lookups)
                    i += len(rule.input)
            else:
                i += 1
        return glyphids


class GSUBLookup:
    ''' GSUB Lookup Table '''
    def __init__(self, ofst: int, fontfile: FontReader):
        self.ofst = ofst
        self.fontfile = fontfile
        fileptr = self.fontfile.tell()

        RIGHT_TO_LEFT = 0x0001
        IGNORE_BASE_GLYPHS = 0x0002
        IGNORE_LIGATURES = 0x0004
        IGNORE_MARKS = 0x0008
        USE_MARK_FILTERING_SET = 0x0010
        MARK_ATTACHMENT_TYPE_MASK = 0xFF00

        self.fontfile.seek(self.ofst)
        self.type = self.fontfile.readuint16()
        self.fmt = 1
        self.flag = self.fontfile.readuint16()
        subtablecnt = self.fontfile.readuint16()
        self.tableofsts = []
        for i in range(subtablecnt):
            self.tableofsts.append(self.fontfile.readuint16())
        self.markfilterset = None
        if self.flag & USE_MARK_FILTERING_SET:
            self.markfilterset = self.fontfile.readuint16()

        self.subtables: list[LookupSubtable] = []
        for i in range(subtablecnt):
            tblofst = self.ofst+self.tableofsts[i]
            self.fmt = self.fontfile.readuint16(tblofst)
            
            tabletype = self.type
            if self.type == 7:  # Extension Table - turns into another type
                tabletype = self.fontfile.readuint16()
                extofst = self.fontfile.readuint32()
                tblofst += extofst
                self.fmt = self.fontfile.readuint16(tblofst)

            if tabletype == 1:  # Single Substitution
                if self.fmt == 1:
                    self.subtables.append(LookupSingleSub1(tblofst, self.fontfile))
                else:
                    self.subtables.append(LookupSingleSub2(tblofst, self.fontfile))

            elif tabletype == 2:  # Multiple sub
                self.subtables.append(LookupMultipleSub(tblofst, self.fontfile))

            elif tabletype == 3:  # Alternates lookup
                self.subtables.append(LookupAlternate(tblofst, self.fontfile))

            elif tabletype == 4:  # Ligature sub
                self.subtables.append(LookupLigatureSub(tblofst, self.fontfile))

            elif tabletype == 6:  # Chained context sub
                if self.fmt == 1:
                    self.subtables.append(LookupChainedSub1(tblofst, self.fontfile))
                elif self.fmt == 2:
                    self.subtables.append(LookupChainedSub2(tblofst, self.fontfile))
                else:
                    self.subtables.append(LookupChainedSub3(tblofst, self.fontfile))
            else:
                logging.debug('Unimplemented GSUB Lookup Type %s', self.type)
                self.subtables.append(LookupSubtable(self.type, self.fontfile))

        self.fontfile.seek(fileptr)

    def __repr__(self):
        return f'<GSUBLookup Type {self.type}.{self.fmt}>'

    def sub(self, glyphids: list[int], lookups: list[GSUBLookup], name: Optional[str] = None) -> list[int]:
        ''' Apply glyph substitution to list of glyph ids '''
        for subtable in self.subtables:
            glyphids = subtable.sub(glyphids, lookups, name)
        return glyphids


class Gsub:
    ''' Glyph Substitution Table '''
    def __init__(self, ofst: int, fontfile: FontReader):
        self.fontfile = fontfile
        self.ofst = ofst
        self.language = Language()
        self.fontfile.seek(self.ofst)
        vermajor = self.fontfile.readuint16()
        verminor = self.fontfile.readuint16()
        scriptofst = self.fontfile.readuint16()
        featureofst = self.fontfile.readuint16()
        lookupofst = self.fontfile.readuint16()
        featurevariationsofst = None
        if verminor == 1:
            featurevariationsofst = self.fontfile.readuint32()
            logging.warning('GSUB has feature variations - unimplemented')

        # Read scripts
        scriptlisttableloc = self.ofst + scriptofst
        scriptcnt = self.fontfile.readuint16(scriptlisttableloc)
        self.scripts = {}
        for i in range(scriptcnt):
            tag = self.fontfile.read(4).decode()
            self.scripts[tag] = Script(
                tag,
                self.fontfile.readuint16() + scriptlisttableloc,
                self.fontfile)

        # Read features
        featurelisttableloc = self.ofst + featureofst
        featurecnt = self.fontfile.readuint16(featurelisttableloc)
        self.featurelist = []
        for i in range(featurecnt):
            self.featurelist.append(Feature(
                self.fontfile.read(4).decode(),
                self.fontfile.readuint16() + featurelisttableloc,
                self.fontfile))

        # Read Lookups
        lookuplisttableloc = self.ofst + lookupofst
        lookupcnt = self.fontfile.readuint16(lookuplisttableloc)
        self.lookups = []
        for i in range(lookupcnt):
            self.lookups.append(GSUBLookup(
                self.fontfile.readuint16() + lookuplisttableloc,
                self.fontfile))

        # Put everything in a dictionary for access
        self.features = {}
        for scrname, script in self.scripts.items():
            langdict = {}
            for langname, lang in script.languages.items():
                langfeatures = [self.featurelist[i] for i in lang.featureidxs]
                featnames = [f.tag for f in langfeatures]
                featdict = {}
                for feat in featnames:
                    lookups = langfeatures[featnames.index(feat)].lookupids
                    tables = [self.lookups[i] for i in lookups]
                    featdict[feat] = tables
                langdict[langname] = featdict
            self.features[scrname] = langdict            

    def features_active(self):
        ''' Dictionary of features active in the current script/language system '''
        return self.features.get(self.language.script, {}).get(self.language.language, {})

    def sub(self, glyphids: list[int], features: FontFeatures):
        ''' Apply glyph substitution to list of glyph ids. Features
            enable/disable certain substitutions.
        '''
        feattable = self.features_active()
        
        def apply_feature(name, glyphids):
            tables = feattable.get(name, [])
            for table in tables:
                newglyphids = table.sub(glyphids.copy(), self.lookups, name)
                if newglyphids != glyphids:
                    logging.debug('GSUB applied by feature %s in %s', name, table)
                glyphids = newglyphids
            return glyphids

        basicfeatures = ['locl', 'ccmp', 'rlig']
        altfeatures = ['rclt', 'rand']
        if features.calt:
            altfeatures.append('calt')
        if features.clig:
            altfeatures.append('clig')
        if features.liga:
            altfeatures.append('liga')
        if features.dlig:
            altfeatures.append('dlig')
        if features.c2sc:
            altfeatures.append('c2sc')
        if features.frac:
            altfeatures.append('frac')
        if features.salt:
            altfeatures.append('salt')
        if features.hlig:
            altfeatures.append('hlig')
        if features.zero:
            altfeatures.append('zero')
        if features.ssty:
            altfeatures.append('ssty')

        # Run basic features first, in order defined by font
        for feat in feattable.keys():
            if feat in basicfeatures:
                glyphids = apply_feature(feat, glyphids)

        # Then alernate/optional features, in order defined by font
        for feat in feattable.keys():
            if feat in altfeatures:
                glyphids = apply_feature(feat, glyphids)

        return glyphids



''' Inspect and debug glyphs '''

from __future__ import annotations
from typing import TYPE_CHECKING
import xml.etree.ElementTree as ET

from .svgpath import fmt

if TYPE_CHECKING:
    from .glyph import SimpleGlyph


class InspectGlyph:
    ''' Draw glyph svg with test/debug lines '''
    def __init__(self, glyph: SimpleGlyph, pxwidth: float = 400, pxheight: float = 400):
        self.glyph = glyph
        self.font = self.glyph.font
        self.pxwidth = pxwidth
        self.pxheight = pxheight

    def _repr_svg_(self):
        ''' Jupyter representation '''
        return self.svg()

    def svg(self) -> str:
        ''' Glyph SVG string '''
        return ET.tostring(self.svgxml(), encoding='unicode')

    def svgxml(self) -> ET.Element:
        ''' Glyph svg as XML element tree '''
        # Draw glyph at 1:1 scale in glyph units
        height = self.font.info.layout.ymax - self.font.info.layout.ymin
        width = height * self.pxwidth/self.pxheight

        # vertical positions
        ymargin = height / 20
        height *= 1.1
        bot = height - ymargin
        baseline = height + self.font.info.layout.ymin - ymargin
        descend = baseline - self.font.info.layout.descent
        ascend = baseline - self.font.info.layout.ascent
        top = baseline - self.font.info.layout.ymax

        # horizontal positions
        xadvance = self.glyph.advance()
        x0 = (width - xadvance)/2
        x1 = x0 + xadvance

        # drawing constants
        radius = fmt(width/150)
        txtmargin = fmt(25)
        txtsize = fmt(width/40)
        tick = width/50

        svg = ET.Element('svg')
        svg.set('xmlns', 'http://www.w3.org/2000/svg')
        svg.set('viewBox', f'0 0 {fmt(width)} {fmt(height)}')
        svg.set('width', fmt(self.pxwidth))
        svg.set('height', fmt(self.pxheight))

        # Height lines
        def hline(y, name, color):
            path = ET.SubElement(svg, 'path')
            path.attrib['d'] = f'M 0 {fmt(y)} l {fmt(width)} 0'
            path.attrib['stroke'] = color
            path.set('stroke-width', '1px')
            path.set('opacity', '.5')
            path.set('vector-effect', 'non-scaling-stroke')
            text = ET.SubElement(svg, 'text')
            text.set('x', fmt(txtmargin))
            text.set('y', fmt(y-20))
            text.set('font-size', txtsize)
            text.text = name
        hline(baseline, 'baseline', 'red')
        hline(bot, 'ymin', 'gray')
        hline(descend, 'descend', 'gray')
        hline(ascend, 'ascend', 'gray')
        hline(top, 'ymax', 'gray')

        # Vertical lines
        def vline(x, color):
            path = ET.SubElement(svg, 'path')
            path.attrib['d'] = f'M {fmt(x)} 0 l 0 {fmt(height)}'
            path.attrib['stroke'] = color
            path.set('stroke-width', '1px')
            path.set('opacity', '.5')
            path.set('vector-effect', 'non-scaling-stroke')
        vline(x0, 'gray')
        vline(x1, 'gray')

        # Ticks
        z = ET.SubElement(svg, 'path')
        z.set('d', (f'M {fmt(x0-tick)} {fmt(baseline)} '
                    f'L {fmt(x0)} {fmt(baseline)} {fmt(x0)} {fmt(baseline+tick)}'))
        z.set('stroke', '#444444')
        z.set('stroke-width', '4px')
        z.set('fill', 'none')
        z.set('vector-effect', 'non-scaling-stroke')
        z = ET.SubElement(svg, 'path')
        z.set('d', (f'M {fmt(x1+tick)} {fmt(baseline)} '
                    f'L {fmt(x1)} {fmt(baseline)} {fmt(x1)} {fmt(baseline+tick)}'))
        z.set('stroke', '#444444')
        z.set('stroke-width', '4px')
        z.set('fill', 'none')
        z.set('vector-effect', 'non-scaling-stroke')
        text = ET.SubElement(svg, 'text')
        text.set('x', fmt(x0))
        text.set('y', fmt(baseline+tick+100))
        text.set('font-size', txtsize)
        text.set('text-anchor', 'middle')
        text.set('alignment-baseline', 'top')
        text.text = '0'
        text = ET.SubElement(svg, 'text')
        text.set('x', fmt(x1))
        text.set('y', fmt(baseline+tick+100))
        text.set('font-size', txtsize)
        text.set('text-anchor', 'middle')
        text.set('alignment-baseline', 'top')
        text.text = str(int(xadvance))

        # Glyph Outline
        g = self.glyph.svgpath(x0=x0, y0=baseline,
                               scale_factor=1/self.glyph._points_per_unit)  # type: ignore
        if g is not None:
            g.set('fill', 'lightgray')
            g.set('stroke', 'black')
            g.set('stroke-width', '2px')
            g.set('vector-effect', 'non-scaling-stroke')
            svg.append(g)

        for op in self.glyph.operators:
            points, ctrls = op.points()
            for p, c in zip(points, ctrls):
                circ = ET.SubElement(svg, 'circle')
                circ.set('cx', fmt(x0 + p.x))
                circ.set('cy', fmt(baseline-p.y))
                circ.set('r', radius)
                circ.set('fill', '#393ee3' if c else '#d1211b')
                circ.set('stroke-width', '1px')
                circ.set('vector-effect', 'non-scaling-stroke')
        return svg


class DescribeGlyph:
    ''' Table of Glyph information '''
    def __init__(self, glyph: SimpleGlyph):
        self.glyph = glyph

    def __repr__(self):
        chstr = ', '.join(list(self.glyph.char))
        ordstr = ', '.join(format(ord(k), '04X') for k in list(self.glyph.char))
        r = f'Index: {self.glyph.index}\n'
        r += f'Unicode: {ordstr}\n'
        r += f'Character: {chstr}\n'
        r += f'xmin: {self.glyph.bbox.xmin}\n'
        r += f'xmax: {self.glyph.bbox.xmax}\n'
        r += f'ymin: {self.glyph.bbox.ymin}\n'
        r += f'ymax: {self.glyph.bbox.ymax}\n'
        r += f'Advance: {self.glyph.advance()}\n'
        if hasattr(self.glyph, 'glyphs'):  # Compound
            comps = self.glyph.glyphs.glyphs
            ids = ', '.join(str(c.index) for c in comps)
            r += f'Component ids: {ids}\n'
        return r

    def _repr_html_(self):
        ''' Jupyter representation, HTML table '''
        return self.describe()

    def describe(self):
        ''' HTML table with glyph parameters '''
        chstr = ', '.join(list(self.glyph.char))
        ordstr = ', '.join(format(ord(k), '04X') for k in list(self.glyph.char))
        comprow = ''
        if hasattr(self.glyph, 'glyphs'):  # Compound
            comps = self.glyph.glyphs.glyphs
            ids = ', '.join(str(c.index) for c in comps)
            comprow = f'<tr><td>Component ids</td><td>{ids}</td></tr>'

        h = f'''
        <table>
        <tr><td>Index</td><td>{self.glyph.index}</td></tr>
        <tr><td>Unicode</td><td>{ordstr}</td></tr>
        <tr><td>Character</td><td>{chstr}</td></tr>
        <tr><td>xmin</td><td>{self.glyph.bbox.xmin}</td></tr>
        <tr><td>xmax</td><td>{self.glyph.bbox.xmax}</td></tr>
        <tr><td>ymin</td><td>{self.glyph.bbox.ymin}</td></tr>
        <tr><td>ymax</td><td>{self.glyph.bbox.ymax}</td></tr>
        <tr><td>Advance</td><td>{self.glyph.advance()}</td></tr>
        {comprow}
        </table>
        '''
        return h


class ShowGlyphs:
    ''' Show all glyphs in the font '''
    def __init__(self, font, size: float = 36, pxwidth: float = 800):
        self.font = font
        self.size = size
        self.pxwidth = pxwidth
        self.linespacing = 1.15

    def _repr_svg_(self):
        ''' Jupyter representation '''
        return self.svg()

    def svg(self) -> str:
        ''' Glyph SVG string '''
        return ET.tostring(self.svgxml(), encoding='unicode')

    def svgxml(self) -> ET.Element:
        lineheight = self.size * self.linespacing
        scale = self.size / self.font.info.layout.unitsperem

        svg = ET.Element('svg')
        svg.set('xmlns', 'http://www.w3.org/2000/svg')
        svg.set('width', fmt(self.pxwidth))

        x = 0
        y = lineheight
        for i in range(self.font.info.header.numglyphs):
            glyph = self.font.glyph_fromid(i)
            if x + glyph.advance() * scale > self.pxwidth:
                y += lineheight
                x = 0
            if x == 0 and glyph.bbox.xmin < 0:
                x = glyph.bbox.xmin * scale
            g = glyph.svgpath(x0=x, y0=y, scale_factor=self.size/glyph.DFLT_SIZE_PT)
            if g is not None:
                svg.append(g)
            x += glyph.advance() * scale

        height = y + lineheight/2
        svg.set('viewBox', f'0 0 {fmt(self.pxwidth)} {fmt(height)}')
        svg.set('height', fmt(height))
        return svg


class ShowLookup4:
    ''' Show items in Lookup Table type 4 '''
    def __init__(self, lookup, font, size: float=36, pxwidth: int=400):
        self.lookup = lookup
        self.font = font
        self.size = size
        self.pxwidth = pxwidth
        self.linespacing = 1.25

    def _repr_svg_(self):
        ''' Jupyter representation '''
        return self.svg()

    def svg(self) -> str:
        ''' Glyph SVG string '''
        return ET.tostring(self.svgxml(), encoding='unicode')

    def svgxml(self) -> ET.Element:
        ''' Get SVG as XML tree '''
        lineheight = self.size * self.linespacing
        scale = self.size / self.font.info.layout.unitsperem

        svg = ET.Element('svg')
        svg.set('xmlns', 'http://www.w3.org/2000/svg')
        svg.set('width', fmt(self.pxwidth))

        y = lineheight
        for subtable in self.lookup.subtables:
            assert subtable.covtable.format == 1
            for covidx, startglyph in enumerate(subtable.covtable.glyphs):
                ligset = subtable.ligsets[covidx]
                for nextglyphs, repl in ligset.items():
                    origglyphs = [startglyph] + list(nextglyphs)

                    x = 0
                    for gid in origglyphs:
                        glyph = self.font.glyph_fromid(gid)
                        g = glyph.svgpath(x0=x, y0=y, scale_factor=self.size/glyph.DFLT_SIZE_PT)
                        if g is not None:
                            svg.append(g)
                        x += glyph.advance() * scale
                    x += 20
                    glyph = self.font.glyph('-')
                    svg.append(glyph.svgpath(x0=x, y0=y,
                                             scale_factor=self.size/glyph.DFLT_SIZE_PT))
                    x += glyph.advance() * scale
                    x += 20

                    glyph = self.font.glyph_fromid(repl)
                    g = glyph.svgpath(x0=x, y0=y, scale_factor=self.size/glyph.DFLT_SIZE_PT)
                    g.set('fill', 'red')
                    svg.append(g)
                    y += lineheight

        svg.set('viewBox', f'0 0 {fmt(self.pxwidth)} {fmt(y + lineheight/2)}')
        svg.set('height', fmt(y + lineheight/2))
        return svg



''' Font CMAP table - translate character into a glyph ID '''

from __future__ import annotations
from typing import Sequence, Optional


class Cmap12:
    ''' Format 12 cmap Table '''
    def __init__(self, platform: str, platformid: int, startcodes: Sequence[int],
                 endcodes: Sequence[int], glyphstarts: Sequence[int]):
        self.platform = platform
        self.platformid = platformid
        self.startcodes = startcodes
        self.endcodes = endcodes
        self.glyphstarts = glyphstarts

        self.glyphmap = {}
        for startid, start in enumerate(self.startcodes):
            for i in range(start, self.endcodes[startid]+1):
                self.glyphmap[i] = self.glyphstarts[startid] + (i-start)

        self.glyphmap_r: Optional[dict[int, set[str]]] = None  # Build on demand

    def __repr__(self):
        return f'<Cmap Format 12: {self.platform} id={self.platformid}'

    def glyphid(self, char: str) -> int:
        ''' Get glyph index from a character '''
        charid = ord(char)
        return self.glyphmap.get(charid, 0)       

    def char(self, glyphid: int) -> set[str]:
        ''' Get unicode character code from glyphid (reverse cmap lookup) '''
        if self.glyphmap_r is None:
            self.glyphmap_r = {}
            for key, val in self.glyphmap.items():
                if val not in self.glyphmap_r:
                    self.glyphmap_r[val] = set([chr(key)])
                else:
                    self.glyphmap_r[val].add(chr(key))
        return self.glyphmap_r.get(glyphid, set())


class Cmap4:
    ''' Format 4 cmap Table '''
    def __init__(self, platform: str, platformid: int, startcodes: Sequence[int],
                 endcodes: Sequence[int], idrangeoffset: Sequence[int],
                 iddeltas: Sequence[int], glyphidarray: Sequence[int]):
        self.platform = platform
        self.platformid = platformid
        self.startcodes = startcodes
        self.endcodes = endcodes
        self.idrangeoffset = idrangeoffset
        self.iddeltas = iddeltas
        self.glyphidarray = glyphidarray

        self.glyphmap = {}
        for seg, startcode in enumerate(self.startcodes):
            for i in range(startcode, self.endcodes[seg]+1):
                if self.idrangeoffset[seg] != 0:
                    idx = (seg 
                           - len(self.startcodes)
                           + self.idrangeoffset[seg]//2
                           + i - self.startcodes[seg])
                    gid = self.glyphidarray[idx]
                    gid = self.iddeltas[seg] + gid if gid > 0 else 0
                else:
                    gid = (self.iddeltas[seg] + i) % 0x10000
                self.glyphmap[i] = gid

        self.glyphmap_r: Optional[dict[int, set[str]]] = None  # Build on demand

    def __repr__(self):
        return f'<Cmap Format 4: {self.platform} id={self.platformid}'

    def glyphid(self, char: str) -> int:
        ''' Get glyph index from a character '''
        charid = ord(char)
        return self.glyphmap.get(charid, 0)       

    def char(self, glyphid: int) -> set[str]:
        ''' Get unicode character code from glyphid (reverse cmap lookup) '''
        if self.glyphmap_r is None:
            self.glyphmap_r = {}
            for key, val in self.glyphmap.items():
                if val not in self.glyphmap_r:
                    self.glyphmap_r[val] = set([chr(key)])
                else:
                    self.glyphmap_r[val].add(chr(key))
        return self.glyphmap_r.get(glyphid, set())
